import logging

logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())


class ResourceBase(object):
    """Base class for JSONAPI resource methods."""

    def __init__(self, session):
        self.session = session
        self.params = {}

    def _get_records(self, entity_type, bundle=None, resource_id=None, params=None):
        """Helper function that checks to retrieve one record, one page or multiple pages of farmOS records"""
        if params is None:
            params = {}

        params = {**self.params, **params}

        path = self._get_resource_path(entity_type, bundle, resource_id)

        response = self.session.http_request(path=path, params=params)
        return response.json()

    def get(self, entity_type, bundle=None, params=None):
        return self._get_records(entity_type=entity_type, bundle=bundle, params=params)

    def get_id(self, entity_type, bundle=None, resource_id=None, params=None):
        return self._get_records(
            entity_type=entity_type,
            bundle=bundle,
            params=params,
            resource_id=resource_id,
        )

    def iterate(self, entity_type, bundle=None, params=None):
        response = self._get_records(
            entity_type=entity_type, bundle=bundle, params=params
        )
        more = True
        while more:
            # TODO: Should we merge in the "includes" info here?
            for record in response["data"]:
                yield record
            try:
                next_url = response["links"]["next"]["href"]
                response = self.session._http_request(url=next_url)
                response = response.json()
            except KeyError:
                more = False

    def send(self, entity_type, bundle=None, payload=None):

        # Default to empty payload dict.
        if payload is None:
            payload = {}

        # Set the resource type.
        payload["type"] = self._get_resource_type(entity_type, bundle)
        json_payload = {
            "data": {**payload},
        }
        options = {"json": json_payload}

        # If an ID is included, update the record
        id = payload.pop("id", None)
        if id:
            json_payload["data"]["id"] = id
            logger.debug("Updating record id: of entity type: %s", id, entity_type)
            path = self._get_resource_path(
                entity_type=entity_type, bundle=bundle, record_id=id
            )
            response = self.session.http_request(
                method="PATCH", path=path, options=options
            )
        # If no ID is included, create a new record
        else:
            logger.debug("Creating record of entity type: %s", entity_type)
            path = self._get_resource_path(entity_type=entity_type, bundle=bundle)
            response = self.session.http_request(
                method="POST", path=path, options=options
            )

        # Handle response from POST requests
        if response.status_code == 201:
            logger.debug("Record created.")

        # Handle response from PUT requests
        if response.status_code == 200:
            logger.debug("Record updated.")

        return response.json()

    def delete(self, entity_type, bundle=None, id=None):
        logger.debug("Deleted record id: %s of entity type: %s", id, entity_type)
        path = self._get_resource_path(
            entity_type=entity_type, bundle=bundle, record_id=id
        )
        return self.session.http_request(method="DELETE", path=path)

    @staticmethod
    def _get_resource_path(entity_type, bundle=None, record_id=None):
        """Helper function that builds paths to jsonapi resources."""

        if bundle is None:
            bundle = entity_type

        path = "api/" + entity_type + "/" + bundle

        if record_id:
            path += "/" + str(record_id)

        return path

    @staticmethod
    def _get_resource_type(entity_type, bundle=None):
        """Helper function that builds a JSONAPI resource name."""

        if bundle is None:
            bundle = entity_type

        return entity_type + "--" + bundle


class ResourceHelperBase(object):
    def __init__(self, session, entity_type):
        self.entity_type = entity_type
        self.resource_api = ResourceBase(session=session)

    def get(self, bundle, params=None):
        return self.resource_api.get(
            entity_type=self.entity_type, bundle=bundle, params=params
        )

    def get_id(self, bundle, resource_id, params=None):
        return self.resource_api.get_id(
            entity_type=self.entity_type,
            bundle=bundle,
            resource_id=resource_id,
            params=params,
        )

    def iterate(self, bundle, params=None):
        return self.resource_api.iterate(
            entity_type=self.entity_type, bundle=bundle, params=params
        )

    def send(self, bundle, payload=None):
        return self.resource_api.send(
            entity_type=self.entity_type, bundle=bundle, payload=payload
        )

    def delete(self, bundle, id):
        return self.resource_api.delete(
            entity_type=self.entity_type, bundle=bundle, id=id
        )


class AssetAPI(ResourceHelperBase):
    """API for interacting with farm assets"""

    def __init__(self, session):
        # Define 'asset' as the JSONAPI resource type.
        super().__init__(session=session, entity_type="asset")


class LogAPI(ResourceHelperBase):
    """API for interacting with farm logs"""

    def __init__(self, session):
        # Define 'log' as the JSONAPI resource type.
        super().__init__(session=session, entity_type="log")


class TermAPI(ResourceHelperBase):
    """API for interacting with farm Terms"""

    def __init__(self, session):
        # Define 'taxonomy_term' as the farmOS API entity endpoint
        super().__init__(session=session, entity_type="taxonomy_term")


def info(session):
    """Retrieve info about the farmOS server."""

    logger.debug("Retrieving farmOS server info.")
    response = session.http_request("api")
    return response.json()


def filter(path, value, operation="="):
    """Helper method to build JSONAPI filter query params."""

    # TODO: Validate path, value, operation?
    # TODO: Support filter groups.
    # TODO: Support revision filters, are these edge cases?
    # TODO: Support "pagination" query params? Separate method?
    # TODO: Support "sort" query params? Separate method?
    # TODO: Support "includes" query params? Separate method?

    filters = {}

    # If the operation is '=', only one query param is required.
    if operation == "=":
        param = f"filter[{path}]"
        filters[param] = value
    # Else we need a query param for the path, operation, and value.
    else:
        # TODO: Use a unique identifier instead of using "condition"
        # otherwise the same path cannot be filtered multiple times.
        base_param = f"filter[{path}_{operation.lower()}][condition]"

        path_param = base_param + "[path]"
        filters[path_param] = path

        op_param = base_param + "[operator]"
        filters[op_param] = operation

        value_param = base_param + "[value]"
        if operation in ["IN", "NOT IN", ">", "<", "<>", "BETWEEN"]:
            value_param += "[]"
        filters[value_param] = value

    return filters



import pytest
import os

import farmOS

# Allow testing via http.
os.environ["OAUTHLIB_INSECURE_TRANSPORT"] = "1"

FARMOS_HOSTNAME = os.getenv("FARMOS_HOSTNAME")

# Credentials for OAuth Auth
FARMOS_OAUTH_USERNAME = os.getenv("FARMOS_OAUTH_USERNAME")
FARMOS_OAUTH_PASSWORD = os.getenv("FARMOS_OAUTH_PASSWORD")
FARMOS_OAUTH_CLIENT_ID = os.getenv("FARMOS_OAUTH_CLIENT_ID", "farm")
FARMOS_OAUTH_CLIENT_SECRET = os.getenv("FARMOS_OAUTH_CLIENT_SECRET", None)

valid_oauth_config = (
    FARMOS_HOSTNAME is not None
    and FARMOS_OAUTH_USERNAME is not None
    and FARMOS_OAUTH_PASSWORD is not None
)

farmOS_testing_server = pytest.mark.skipif(
    not valid_oauth_config,
    reason="farmOS Testing Server not configured. Skipping farmOS test server integration tests.",
)


@pytest.fixture(scope="module")
def test_farm():
    if valid_oauth_config:
        farm = farmOS.farmOS(
            hostname=FARMOS_HOSTNAME,
            scope="farm_manager",
            client_id=FARMOS_OAUTH_CLIENT_ID,
            client_secret=FARMOS_OAUTH_CLIENT_SECRET,
            version=2,
        )
        farm.authorize(username=FARMOS_OAUTH_USERNAME, password=FARMOS_OAUTH_PASSWORD)
        return farm


@pytest.fixture(scope="module")
def test_logs(test_farm):
    log_ids = []
    # Create logs.
    for x in range(1, 60):
        test_log = {
            "type": "observation",
            "payload": {"attributes": {"name": "Log #" + str(x)}},
        }
        response = test_farm.log.send(test_log["type"], test_log["payload"])
        log_ids.append(response["data"]["id"])

    return log_ids


@pytest.fixture(scope="module")
def test_assets(test_farm):
    asset_ids = []
    # Create assets.
    for x in range(1, 60):
        test_asset = {
            "type": "equipment",
            "payload": {"attributes": {"name": "Asset #" + str(x)}},
        }
        response = test_farm.asset.send(test_asset["type"], test_asset["payload"])
        asset_ids.append(response["data"]["id"])

    return asset_ids


@pytest.fixture(scope="module")
def test_terms(test_farm):
    term_ids = []
    # Create terms.
    for x in range(1, 60):
        test_term = {
            "type": "plant_type",
            "payload": {"attributes": {"name": "Plant type #" + str(x)}},
        }
        response = test_farm.term.send(test_term["type"], test_term["payload"])
        term_ids.append(response["data"]["id"])

    return term_ids



import logging
from urllib.parse import urlparse, parse_qs

logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())


class BaseAPI(object):
    """Base class for API methods

    This class includes basic implementations of the
    get(), send() and delete() methods used in the API client.

    Keyword Arguments:
    session: APISession for the farmOS instance
    entity_type: String, used to set the entity type in the path
        of all requests used with the API
    """

    def __init__(self, session, entity_type=None):
        self.session = session
        self.entity_type = entity_type
        self.filters = {}

    def _get_single_record_data(self, id):
        """Retrieve one record given the record ID"""
        # Set path to return record type by specific ID
        path = self.entity_type + "/" + str(id) + ".json"

        logger.debug(
            "Getting single record data for id: %s of entity type: %s",
            id,
            self.entity_type,
        )
        response = self.session.http_request(path=path)

        return response.json()

    def _get_record_data(self, filters):
        """Retrieve one page of raw record data from the farmOS API."""
        # Set path to return record type + filters
        path = self.entity_type + ".json"
        # Combine instance filters and filters from the method call
        filters = {**self.filters, **filters}

        response = self.session.http_request(path=path, params=filters)

        # Return object
        data = {}

        response = response.json()
        if "list" in response:
            data["list"] = response["list"]
            data["page"] = {
                "self": _parse_api_page(url=response["self"]),
                "first": _parse_api_page(url=response["first"]),
                "last": _parse_api_page(url=response["last"]),
            }

        return data

    def _get_all_record_data(self, filters, page=0, list=None):
        """Recursive function to retrieve multiple pages of raw record data from the farmOS API."""
        if list is None:
            list = []

        filters["page"] = page

        logger.debug(
            "Getting page: %s of record data of entity type: %s",
            filters["page"],
            self.entity_type,
        )
        data = self._get_record_data(filters=filters)

        # Append record data to list of all requested data
        if "list" in data:
            list = list + data["list"]

        # Check to see if there are more pages
        if "page" in data:
            last_page = data["page"]["last"]
            # Last page, return the list
            if last_page == page:
                data = {
                    "page": {
                        "first": data["page"]["first"],
                        "last": data["page"]["last"],
                    },
                    "list": list,
                }
                return data
            # Recursive call, get the next page
            else:
                return self._get_all_record_data(
                    list=list, page=(page + 1), filters=filters
                )

    def _get_records(self, filters=None):
        """Helper function that checks to retrieve one record, one page or multiple pages of farmOS records"""
        if filters is None:
            filters = {}

        # Determine if filters is an int (id) or dict (filters object)
        if isinstance(filters, int) or isinstance(filters, str):
            return self._get_single_record_data(filters)
        elif isinstance(filters, dict):
            # Check if the caller requests a specific page
            if "page" in filters:
                logger.debug(
                    "Getting page: %s of record data of entity type: %s",
                    filters["page"],
                    self.entity_type,
                )
                return self._get_record_data(filters=filters)
            else:
                logger.debug(
                    "Getting all record data of entity type: %s", self.entity_type
                )
                return self._get_all_record_data(filters=filters)

    def get(self, filters=None):
        """Simple get method"""

        data = self._get_records(filters=filters)

        return data

    def send(self, payload):
        options = {"json": payload}

        # If an ID is included, update the record
        id = payload.pop("id", None)
        if id:
            logger.debug("Updating record id: of entity type: %s", id, self.entity_type)
            path = self.entity_type + "/" + str(id)
            response = self.session.http_request(
                method="PUT", path=path, options=options
            )
        # If no ID is included, create a new record
        else:
            logger.debug("Creating record of entity type: %s", self.entity_type)
            path = self.entity_type
            response = self.session.http_request(
                method="POST", path=path, options=options
            )

        # Handle response from POST requests
        if response.status_code == 201:
            logger.debug("Record created.")
            return response.json()

        # Handle response from PUT requests
        if response.status_code == 200:
            logger.debug("Record updated.")
            # farmOS returns no response data for PUT requests
            # response_data = response.json()

            # Hard code the entity ID, path, and resource
            entity_data = {
                "id": id,
                "uri": path,
                "resource": self.entity_type,
            }
            return entity_data

    def delete(self, id):
        logger.debug("Deleted record id: %s of entity type: %s", id, self.entity_type)
        path = self.entity_type + "/" + str(id)
        response = self.session.http_request(method="DELETE", path=path)

        return response


class TermAPI(BaseAPI):
    """API for interacting with farm Terms"""

    def __init__(self, session):
        # Define 'taxonomy_term' as the farmOS API entity endpoint
        super().__init__(session=session, entity_type="taxonomy_term")

    def get(self, filters=None):
        """Get method that supports a bundle name as the 'filter' parameter"""

        # Check if filters parameter is a str
        if isinstance(filters, str):
            # Add filters to instance requests.session filter dict with keyword 'bundle'
            self.filters["bundle"] = filters
            # Reset filters to empty dict
            filters = {}

        data = self._get_records(filters=filters)

        return data


class LogAPI(BaseAPI):
    """API for interacting with farm logs"""

    def __init__(self, session):
        # Define 'log' as the farmOS API entity endpoint
        super().__init__(session=session, entity_type="log")


class AssetAPI(BaseAPI):
    """API for interacting with farm assets"""

    def __init__(self, session):
        # Define 'farm_asset' as the farmOS API entity endpoint
        super().__init__(session=session, entity_type="farm_asset")


class AreaAPI(TermAPI):
    """API for interacting with farm areas, a subset of farm terms"""

    def __init__(self, session):
        super().__init__(session=session)
        self.filters["bundle"] = "farm_areas"

    def get(self, filters=None):
        """Retrieve raw record data from the farmOS API.

        Override get() from BaseAPI to support TID (Taxonomy ID)
        rather than record ID
        """

        # Determine if filters is an int (tid) or dict (filters object)
        if isinstance(filters, int) or isinstance(filters, str):
            tid = str(filters)
            # Add tid to filters object
            filters = {
                "tid": tid,
            }

        data = self._get_records(filters=filters)

        return data


def info(session):
    """Retrieve info about the farmOS server."""

    logger.debug("Retrieving farmOS server info.")
    response = session.http_request("farm.json")
    return response.json()


def _parse_api_page(url):
    """Helper function that returns page numbers from the API response.

    The farmOS API returns URLs for the self, first and last pages of records
    in requests that return many records. This function helps parse a raw page
    number from the returned URL.
    """

    parsed_url = urlparse(url)
    page_num = parse_qs(parsed_url.query)["page"][0]

    return int(page_num)



import logging

from requests_oauthlib import OAuth2Session
from oauthlib.oauth2 import LegacyApplicationClient

logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())


class OAuthSession(OAuth2Session):
    """OAuthSession uses OAuth2 to authenticate with farmOS"""

    def __init__(
        self,
        hostname,
        client_id,
        client_secret=None,
        scope=None,
        token=None,
        token_url=None,
        authorization_url=None,
        token_updater=None,
        *args,
        **kwargs
    ):

        # Default to the "user_access" scope.
        if scope is None:
            scope = "user_access"

        # Create a dictionary of credentials required to pass along with Refresh Tokens
        # Required to generate a new access token
        auto_refresh_kwargs = {"client_id": client_id, "client_secret": client_secret}

        super(OAuthSession, self).__init__(
            token=token,
            client=LegacyApplicationClient(client_id=client_id),
            auto_refresh_url=token_url,
            auto_refresh_kwargs=auto_refresh_kwargs,
            token_updater=token_updater,
            scope=scope,
        )

        # Save values for later use.
        self._token_url = token_url
        self._authorization_base_url = authorization_url
        self._client_id = client_id
        self._client_secret = client_secret
        self.hostname = hostname

    def authorize(self, username, password, scope=None):
        """Authorize with the farmOS OAuth server."""

        token = self.token

        logger.debug("Retrieving new OAuth Token.")

        # Ask for username if not provided.
        if username is None:
            from getpass import getpass

            username = getpass("Enter username: ")

        # Ask for password if not provided.
        if password is None:
            from getpass import getpass

            password = getpass("Enter password: ")

        # Use default scope if none provided.
        if scope is None:
            scope = self.scope

        token = self.fetch_token(
            token_url=self._token_url,
            client_id=self._client_id,
            client_secret=self._client_secret,
            username=username,
            password=password,
            scope=scope,
        )

        logger.debug("Fetched OAuth Access Token %s", token)

        # Save the token.
        logger.debug("Saving token with token_updater utility.")
        if self.token_updater is not None:
            self.token_updater(token)

        return token

    def http_request(self, path, method="GET", options=None, params=None, headers=None):
        """Raw HTTP request helper function.

        Keyword arguments:
        :param path: the URL path.
        :param method: the HTTP method.
        :param options: a dictionary of data and parameters to pass on to the request.
        :param params: URL query parameters.
        :return: requests response object.
        """

        # Strip protocol, hostname, leading/trailing slashes, and whitespace from the path.
        path = path.strip("/")
        path = path.strip()

        # Assemble the URL.
        url = "{}/{}".format(self.hostname, path)
        return self._http_request(
            url=url, method=method, options=options, params=params, headers=headers
        )

    def _http_request(self, url, method="GET", options=None, params=None, headers=None):

        # Automatically follow redirects, unless this is a POST request.
        # The Python requests library converts POST to GET during a redirect.
        # Allow this to be overridden in options.
        allow_redirects = True
        if method in ["POST", "PUT"]:
            allow_redirects = False
        if options and "allow_redirects" in options:
            allow_redirects = options["allow_redirects"]

        if headers is None:
            headers = {}

        # If there is data to be sent, include it.
        data = None
        if options and "data" in options:
            data = options["data"]
            headers["Content-Type"] = "application/vnd.api+json"

        # If there is a json data to be sent, include it.
        json = None
        if options and "json" in options:
            json = options["json"]
            if "Content-Type" not in headers:
                headers["Content-Type"] = "application/vnd.api+json"

        # Perform the request.
        response = self.request(
            method,
            url,
            headers=headers,
            allow_redirects=allow_redirects,
            data=data,
            json=json,
            params=params,
        )

        # If this is a POST request, and a redirect occurred, attempt to re-POST.
        redirect_codes = [300, 301, 302, 303, 304, 305, 306, 307, 308]
        if method in ["POST", "PUT"] and response.status_code in redirect_codes:
            if response.headers["Location"]:
                response = self.request(
                    method,
                    response.headers["Location"],
                    allow_redirects=True,
                    data=data,
                    json=json,
                    params=params,
                )

        # Raise exception if error.
        response.raise_for_status()

        # Return the response.
        return response



from enum import Enum
from typing import Any, Dict, List, Optional, Union
import json

from pydantic import BaseModel, Field, validator

from .session import OAuthSession

# Subrequests model derived from provided JSON Schema
# https://git.drupalcode.org/project/subrequests/-/blob/3.x/schema.json


class Action(Enum):
    view = "view"
    create = "create"
    update = "update"
    replace = "replace"
    delete = "delete"
    exists = "exists"
    discover = "discover"
    noop = "noop"


class Subrequest(BaseModel):
    action: Action = Field(
        ...,
        description="The action intended for the request. Each action can resolve into a different HTTP method.",
        title="Action",
    )
    endpoint: Optional[str] = Field(
        None,
        description="The API endpoint to request. The base URL will be added automatically.",
        title="Endpoint",
    )
    uri: Optional[str] = Field(
        None, description="The URI where to make the subrequest.", title="URI"
    )
    requestId: Optional[str] = Field(
        None,
        description="ID other requests can use to reference this request.",
        title="Request ID",
    )
    body: Optional[Union[str, dict]] = Field(
        None,
        description="The JSON encoded body payload for HTTP requests send a body.",
        title="Body",
    )
    headers: Optional[Dict[str, Any]] = Field(
        {}, description="HTTP headers to be sent with the request.", title="Headers"
    )
    waitFor: Optional[List[str]] = Field(
        None,
        description="ID of other requests that this request depends on.",
        title="Parent ID",
    )

    @validator("uri", pre=True, always=True)
    def check_uri_or_endpoint(cls, uri, values):
        endpoint = values.get("endpoint", None)
        if uri is None and endpoint is None:
            raise ValueError("Either uri or endpoint is required.")
        return uri

    @validator("body", pre=True)
    def serialize_body(cls, body):
        if not isinstance(body, str):
            return json.dumps(body)
        return body


class SubrequestsBlueprint(BaseModel):
    __root__: List[Subrequest] = Field(
        ...,
        description="Describes the subrequests payload format.",
        title="Subrequests format",
    )


class Format(str, Enum):
    html = "html"
    json = "json"


class SubrequestsBase(object):
    """Class for handling subrequests"""

    subrequest_path = "subrequests"

    def __init__(self, session: OAuthSession):
        self.session = session

    def send(
        self,
        blueprint: Union[SubrequestsBlueprint, List],
        format: Optional[Union[Format, str]] = Format.json,
    ):

        if isinstance(blueprint, List):
            blueprint = SubrequestsBlueprint.parse_obj(blueprint)

        # Modify each sub-request as needed.
        for sub in blueprint.__root__:
            # Build the URI if an endpoint is provided.
            if sub.uri is None and sub.endpoint is not None:
                sub.uri = "{}/{}".format(self.session.hostname, sub.endpoint)

            # Auto populate headers for each sub-request.
            if "Accept" not in sub.headers:
                sub.headers["Accept"] = "application/vnd.api+json"
            if sub.body is not None and "Content-Type" not in sub.headers:
                sub.headers["Content-Type"] = "application/vnd.api+json"

        headers = {"Content-Type": "application/json"}

        params = {}
        if format == Format.json.value:
            params = {"_format": "json"}

        # Generate the json to send. It is important to use the .json() method
        # of the model for correct serialization.
        json = blueprint.json(exclude={"subrequest": {"endpoint"}}, exclude_none=True)
        options = {"data": json}

        response = self.session.http_request(
            method="POST",
            path=self.subrequest_path,
            options=options,
            params=params,
            headers=headers,
        )

        # Return a json response if requested.
        if format == Format.json.value:
            return response.json()

        return response



import setuptools

setup_requires = [
    "pytest-runner",
]

install_requires = ["requests-oauthlib", "pydantic>=1.7.3"]

tests_require = ["pytest", "black>=20"]

with open("README.md", "r") as fh:
    readme = fh.read()

setuptools.setup(
    name="farmOS",
    version="1.0.0-alpha.2",
    author="farmOS team",
    author_email="mike@mstenta.net",
    description="A Python library for interacting with farmOS over API. ",
    long_description=readme,
    long_description_content_type="text/markdown",
    url="https://github.com/farmOS/farmOS.py",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3",
    setup_requires=setup_requires,
    install_requires=install_requires,
    tests_require=tests_require,
    test_suite="pytest",
)



import logging
from datetime import datetime
from urllib.parse import urlparse, urlunparse
from functools import partial

from .session import OAuthSession

from . import client
from . import client_2
from . import subrequests

logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())


class farmOS:
    """A client that connects to the farmOS server."""

    def __init__(
        self,
        hostname,
        client_id="farm",
        client_secret=None,
        scope="user_access",
        token=None,
        token_updater=None,
        version="1",
    ):
        """
        Initialize instance of the farmOS client that connects to a single farmOS server.

        :param hostname: Valid hostname without a path or query params. The HTTPS scheme
            will be added if none is specified.
        :param client_id: OAuth Client ID. Defaults to "farm"
        :param client_secret: OAuth Client Secret. Defaults to None.
        :param scope: OAuth Scope. Defaults to "user_access".
        :param token: An existing OAuth token to use.
        :param token_updater: A function used to save OAuth tokens outside of the client.
        """

        logger.debug("Creating farmOS client.")

        # Load the token updater function.
        self.token_updater = None
        if token_updater is not None:
            logger.debug("Using provided token_updater utility.")
            self.token_updater = token_updater

        self.session = None

        if hostname is not None:
            valid_schemes = ["http", "https"]
            default_scheme = "https"
            parsed_url = urlparse(hostname)

            # Validate the hostname.
            # Add a default scheme if not provided.
            if not parsed_url.scheme:
                parsed_url = parsed_url._replace(scheme=default_scheme)
                logger.debug("No scheme provided. Using %s", default_scheme)

            # Check for a valid scheme.
            if parsed_url.scheme not in valid_schemes:
                raise Exception("Not a valid scheme.")

            # If not netloc was provided, it was probably parsed as the path.
            if not parsed_url.netloc and parsed_url.path:
                parsed_url = parsed_url._replace(netloc=parsed_url.path)
                parsed_url = parsed_url._replace(path="")

            # Check for netloc.
            if not parsed_url.netloc:
                raise Exception("Invalid hostname. Must have netloc.")

            # Don't allow path, params, or query.
            if parsed_url.path or parsed_url.params or parsed_url.query:
                raise Exception("Hostname cannot include path and query parameters.")

            # Build the url again to include changes.
            hostname = urlunparse(parsed_url)
            logger.debug("Complete hostname configured as %s", hostname)

        else:
            raise Exception("No hostname provided and could not be loaded from config.")

        logger.debug("Creating an OAuth Session.")
        # OR implement a method to check both token paths.
        # maybe version can default to none, and check the server?
        token_url = hostname + "/oauth/token"

        # Check the token expiration time.
        if token is not None and "expires_at" in token:
            # Create datetime objects for comparison.
            now = datetime.now()
            expiration_time = datetime.fromtimestamp(float(token["expires_at"]))

            # Calculate seconds until expiration.
            timedelta = expiration_time - now
            expires_in = timedelta.total_seconds()

            # Update the token expires_in value
            token["expires_in"] = expires_in

            # Unset the 'expires_at' key.
            token.pop("expires_at")

        # Create an OAuth Session
        self.session = OAuthSession(
            hostname=hostname,
            client_id=client_id,
            client_secret=client_secret,
            scope=scope,
            token=token,
            token_url=token_url,
            token_updater=self.token_updater,
        )

        self._client_id = client_id
        self._client_secret = client_secret

        if self.session is None:
            raise Exception(
                "Could not create a session object. Supply authentication credentials when "
                "initializing a farmOS Client."
            )

        if version == 2:
            self.log = client_2.LogAPI(self.session)
            self.asset = client_2.AssetAPI(self.session)
            self.term = client_2.TermAPI(self.session)
            self.resource = client_2.ResourceBase(self.session)
            self.info = partial(client_2.info, self.session)
            self.subrequests = subrequests.SubrequestsBase(self.session)
            self.filter = client_2.filter
        else:
            self.log = client.LogAPI(self.session)
            self.asset = client.AssetAPI(self.session)
            self.area = client.AreaAPI(self.session)
            self.term = client.TermAPI(self.session)
            self.info = partial(client.info, self.session)

    def authorize(self, username=None, password=None, scope=None):
        """Authorize with the farmOS server.

        The client must be authorized with the farmOS server before making requests.
        This method utilizes the OAuth Password Credentials flow to authorize users.

        :param username: farmOS Username. Prompted if not included.
        :param password: farmOS Password. Prompted if not included.
        :param scope: Scope to authorize as with the farmOS server. Defaults to "user_access".
        :return: OAuth Token.
        """

        return self.session.authorize(username, password, scope)
